/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/

#include "conference.h"

cnu con_user_new(cnr room, jid id)
{
    pool p;
    cnu user;
    char *key;

    log_debug(NAME, "[%s] adding user %s to room %s", FZONE, jid_full(id), jid_full(room->id));

    p = pool_new();
    user = pmalloco(p, sizeof(_cnu));
    user->p = p;
    user->realid = jid_new(p, jid_full(id));
    user->room = room;
    user->presence = jutil_presnew(JPACKET__AVAILABLE, NULL, NULL);

    key = j_strdup(jid_full(user->realid));
    xhash_put(room->remote, key, (void*)user);

    /* Add this user to the room roster */
    add_roster(room, user->realid);

    /* If admin, switch */
    if(is_admin(room, user->realid) && !is_moderator(room, user->realid))
    {
	log_debug(NAME, "[%s] Adding %s to moderator list", FZONE, jid_full(user->realid));

	/* Update affiliate info */
	add_affiliate(room->admin, user->realid, NULL);
	add_role(room->moderator, user);
    }
    else if(is_member(room, user->realid) && !is_admin(room, user->realid))
    {
	/* Update affiliate information */
	add_affiliate(room->member, user->realid, NULL);
    }
    else if(room->moderated == 1 && room->defaulttype == 1)
    {
        /* Auto-add to participant list if moderated and member type is default */
	add_role(room->participant, user);
    }

    return user;
}

void _con_user_history_send(cnu to, xmlnode node)
{

    if(to == NULL || node == NULL)
    {
        return;
    }

    xmlnode_put_attrib(node, "to", jid_full(to->realid));
    deliver(dpacket_new(node), NULL);
}

void _con_user_nick(xht h, const char *key, void *data, void *arg)
{
    cnu to = (cnu)data;
    cnu from = (cnu)arg;
    char *old, *status, *reason, *actor;
    xmlnode node;
    xmlnode element;
    jid fullid;

    /* send old style unavail pres w/ old nick */
    if((old = xmlnode_get_attrib(from->nick,"old")) != NULL)
    {
	
        node = jutil_presnew(JPACKET__UNAVAILABLE, jid_full(to->realid), NULL);

        fullid = jid_new(xmlnode_pool(node), jid_full(from->localid));
        jid_set(fullid, old, JID_RESOURCE);
        xmlnode_put_attrib(node, "from", jid_full(fullid));

	status = xmlnode_get_attrib(from->nick,"status");
	log_debug(NAME, "[%s] status = %s", FZONE, status);

	reason = xmlnode_get_attrib(from->nick,"reason");
	actor = xmlnode_get_attrib(from->nick,"actor");
	
	if(xmlnode_get_data(from->nick) != NULL)
		node = add_extended_presence(from, to, node, STATUS_MUC_NICKCHANGE, NULL, NULL);
	else if(status || reason || actor)
		node = add_extended_presence(from, to, node, status, reason, actor);

	deliver(dpacket_new(node), NULL);
    }

    /* if there's a new nick, broadcast that too */
    if(xmlnode_get_data(from->nick) != NULL)
    {
	status = xmlnode_get_attrib(from->nick,"status");
	log_debug(NAME, "[%s] status = %s/%s", FZONE, status, STATUS_MUC_CREATED);

	if(j_strcmp(status, STATUS_MUC_CREATED) == 0)
	    node = xmlnode_dup(add_extended_presence(from, to, NULL, status, NULL, NULL));
	else
	    node = xmlnode_dup(add_extended_presence(from, to, NULL, NULL, NULL, NULL));

	/* Hide x:delay, not needed */
	element = xmlnode_get_tag(node, "x?xmlns=jabber:x:delay");
	if(element)
	    xmlnode_hide(element);

        xmlnode_put_attrib(node, "to", jid_full(to->realid));

        fullid = jid_new(xmlnode_pool(node), jid_full(from->localid));
        jid_set(fullid,xmlnode_get_data(from->nick), JID_RESOURCE);
        xmlnode_put_attrib(node, "from", jid_full(fullid));

        deliver(dpacket_new(node), NULL);
    }
}

void con_user_nick(cnu user, char *nick, xmlnode data)
{
    xmlnode node;
    char *status, *reason, *actor;
    cnr room = user->room;

    log_debug(NAME, "[%s] in room %s changing nick for user %s to %s from %s", FZONE, jid_full(room->id), jid_full(user->realid), nick, xmlnode_get_data(user->nick));

    node = xmlnode_new_tag("n");
    xmlnode_put_attrib(node, "old", xmlnode_get_data(user->nick));

    if (data)
    {
	status = xmlnode_get_attrib(data, "status");
	reason = xmlnode_get_data(data);
	actor = xmlnode_get_attrib(data, "actor");

        if(status)
            xmlnode_put_attrib(node, "status", status);

        if(reason)
	    xmlnode_put_attrib(node, "reason", reason);

	if(actor)
	    xmlnode_put_attrib(node, "actor", actor);

	log_debug(NAME, "[%s] status = %s", FZONE, status);
    }

    xmlnode_insert_cdata(node,nick,-1);
    xmlnode_free(user->nick);

    user->nick = node;
    xhash_walk(room->local, _con_user_nick, (void*)user);

    /* send nick change notice if availble */
    if(room->note_rename != NULL && nick != NULL && xmlnode_get_attrib(node, "old") != NULL && j_strlen(room->note_rename) > 0 && user->legacy == 1)
        con_room_send(room, jutil_msgnew("groupchat", NULL, NULL, spools(user->p, xmlnode_get_attrib(node, "old"), " ", room->note_rename, " ", nick, user->p)));
}

void _con_user_enter(xht h, const char *key, void *data, void *arg)
{
    cnu from = (cnu)data;
    cnu to = (cnu)arg;
    xmlnode node;
    jid fullid;

    /* mirror */
    if(from == to)
        return;

    node = xmlnode_dup(add_extended_presence(from, to, NULL, NULL, NULL, NULL));

    xmlnode_put_attrib(node, "to", jid_full(to->realid));

    fullid = jid_new(xmlnode_pool(node), jid_full(from->localid));
    jid_set(fullid, xmlnode_get_data(from->nick), JID_RESOURCE);
    xmlnode_put_attrib(node,"from",jid_full(fullid));

    deliver(dpacket_new(node),NULL);
}

void con_user_enter(cnu user, char *nick, int created)
{
    xmlnode node;
    xmlnode message;
    char *key;
    int h, tflag = 0;
    cnr room = user->room;

    user->localid = jid_new(user->p, jid_full(room->id));
    jid_set(user->localid, shahash(jid_full(user->realid)), JID_RESOURCE);

    key = j_strdup(user->localid->resource);
    xhash_put(room->local, key, (void*)user);

    room->count++;

    log_debug(NAME, "[%s] officiating user %s in room (created = %d) %s as %s/%s", FZONE, jid_full(user->realid), created, jid_full(room->id), nick, user->localid->resource);

    if(j_strlen(room->description) > 0)
    {
        message = jutil_msgnew("groupchat", jid_full(user->realid), NULL, room->description);
	xmlnode_put_attrib(message,"from", jid_full(room->id));
	deliver(dpacket_new(message), NULL);
    }

    if(user->legacy > 0)
    {
        message = jutil_msgnew("groupchat", jid_full(user->realid), NULL, spools(user->p, "This room supports the MUC protocol.", user->p));
	xmlnode_put_attrib(message,"from", jid_full(room->id));
	deliver(dpacket_new(message), NULL);
    }

    if(room->locked > 0)
    {
        message = jutil_msgnew("groupchat", jid_full(user->realid), NULL, spools(user->p, "This room is locked from entry until configuration is confirmed.", user->p));
	xmlnode_put_attrib(message,"from", jid_full(room->id));
	deliver(dpacket_new(message), NULL);
    }

    xhash_walk(room->local, _con_user_enter, (void*)user);

    if(created == 1)
    {
        node = xmlnode_new_tag("reason");
	xmlnode_put_attrib(node, "status", STATUS_MUC_CREATED);
        con_user_nick(user, nick, node); /* pushes to everyone (including ourselves) our entrance */
    }
    else
    {
        con_user_nick(user, nick, NULL); /* pushes to everyone (including ourselves) our entrance */
    }

    /* loop through history and send back */
    if(room->master->history == 0)
        return;

    h = room->hlast;
    while(1)
    {
        h++;

        if(h == room->master->history)
            h = 0;

        _con_user_history_send(user, xmlnode_dup(room->history[h]));

        if(xmlnode_get_tag(room->history[h],"subject") != NULL)
            tflag = 1;

        if(h == room->hlast)
            break;
    }

    /* send last know topic */
    if(tflag == 0 && room->topic != NULL)
    {
        node = jutil_msgnew("groupchat", jid_full(user->realid), xmlnode_get_attrib(room->topic,"subject"), xmlnode_get_data(room->topic));
        xmlnode_put_attrib(node, "from", jid_full(room->id));
        deliver(dpacket_new(node), NULL);
    }

    /* send entrance notice if available */
    if(room->note_join != NULL && j_strlen(room->note_join) > 0 && user->legacy == 1)
        con_room_send(room, jutil_msgnew("groupchat", NULL, NULL, spools(user->p, nick, " ", room->note_join, user->p)));

    /* Send 'non-anonymous' message if necessary */
    if(room->visible == 1)
        con_send_alert(user, NULL, NULL, STATUS_MUC_SHOWN_JID);
}


void con_user_process(cnu to, cnu from, jpacket jp)
{
    xmlnode node, element;
    cnr room = to->room;
    char str[10];
    int t;

    /* we handle all iq's for this id, it's *our* id */
    if(jp->type == JPACKET_IQ)
    {
        if(NSCHECK(jp->iq,NS_BROWSE))
        {
            jutil_iqresult(jp->x);
            node = xmlnode_insert_tag(jp->x, "user");
            xmlnode_put_attrib(node, "xmlns", NS_BROWSE);
            xmlnode_put_attrib(node, "name", xmlnode_get_data(to->nick));

            if(room->visible == 1)
            {
                element = xmlnode_insert_tag(node, "user");
                xmlnode_put_attrib(element, "jid", jid_full(to->realid));
            }
	   
	    if(to->legacy)
                xmlnode_insert_cdata(xmlnode_insert_tag(node, "ns"), NS_CONFERENCE, -1);
	    else
                xmlnode_insert_cdata(xmlnode_insert_tag(node, "ns"), NS_MUC, -1);

            deliver(dpacket_new(jp->x), NULL);
            return;
        }

        if(NSCHECK(jp->iq,NS_LAST))
        {
            jutil_iqresult(jp->x);

            node = xmlnode_insert_tag(jp->x, "query");
            xmlnode_put_attrib(node, "xmlns", NS_LAST);

            t = time(NULL) - to->last;
            sprintf(str,"%d",t);

            xmlnode_put_attrib(node ,"seconds", str);

            deliver(dpacket_new(jp->x), NULL);
            return;
        }

        /* deny any other iq's if it's private */
        if(to->private == 1)
        {
            jutil_error(jp->x, TERROR_FORBIDDEN);
            deliver(dpacket_new(jp->x), NULL);
            return;
        }

        /* if not, fall through and just forward em on I guess! */
    }

    /* Block possibly faked groupchat messages - groupchat is not meant for p2p chats */
    if(jp->type == JPACKET_MESSAGE)
    {
        if(jp->subtype == JPACKET__GROUPCHAT)
	{
	    jutil_error(jp->x, TERROR_BAD);
	    deliver(dpacket_new(jp->x), NULL);
	    return;
	}

	if(room->privmsg == 1 && !is_admin(room, from->realid))
	{
	    /* Only error on messages with body, otherwise just drop */
	    if(xmlnode_get_tag(jp->x, "body") != NULL)
            {	
	        jutil_error(jp->x, TERROR_MUC_PRIVMSG);
	        deliver(dpacket_new(jp->x), NULL);
	        return;
	    }
	    else
	    {
		xmlnode_free(jp->x);
		return;
	    }
	}
	
    }

    con_user_send(to, from, jp->x);
}

void con_user_send(cnu to, cnu from, xmlnode node)
{
    jid fullid;

    if(to == NULL || from == NULL || node == NULL)
    {
        return;
    }

    fullid = jid_new(xmlnode_pool(node), jid_full(from->localid));
    xmlnode_put_attrib(node, "to", jid_full(to->realid));

    if(xmlnode_get_attrib(node, "cnu") != NULL)
        xmlnode_hide_attrib(node, "cnu");

    jid_set(fullid, xmlnode_get_data(from->nick), JID_RESOURCE);

    xmlnode_put_attrib(node, "from", jid_full(fullid));
    deliver(dpacket_new(node), NULL);
}

void con_user_zap(cnu user, xmlnode data)
{
    cnr room;
    char *reason;
    char *status;

    if(user == NULL)
        return;

    user->leaving = 1;
    status = xmlnode_get_attrib(data, "status");
    reason = xmlnode_get_data(data);

    room = user->room;
    log_debug(NAME, "[%s] zapping user %s <%s-%s>", FZONE, jid_full(user->realid), status, reason);

    if(user->localid != NULL)
    {
        con_user_nick(user, NULL, data); /* sends unavailve */
        xhash_zap(room->local, user->localid->resource);
        room->count--;

        /* send departure notice if available*/
	if(room->note_leave != NULL && j_strlen(room->note_leave) > 0 && user->legacy == 1)
	{
	    if(reason != NULL && j_strcmp(status, STATUS_MUC_BANNED) != 0)
            {
	        con_room_send(room,jutil_msgnew("groupchat",NULL,NULL,spools(user->p,xmlnode_get_attrib(user->nick,"old")," ",room->note_leave," : ", reason, user->p)));
	    }
            else 
	    {
	        con_room_send(room,jutil_msgnew("groupchat",NULL,NULL,spools(user->p,xmlnode_get_attrib(user->nick,"old")," ",room->note_leave,user->p)));
	    }
	}
    }

    remove_affiliate(room->admin, user->realid);
    remove_affiliate(room->member, user->realid);

    revoke_role(room->moderator, user);
    revoke_role(room->participant, user);

    remove_roster(room, user->realid);

    xhash_zap(room->remote,jid_full(user->realid));

    xmlnode_free(user->presence);
    xmlnode_free(user->nick);
    pool_free(user->p);
}

